@@ -48,12 +48,14 @@ def _print_diff(diff, no_color=False) -> None:
         if no_color:
             print(f"{op} event {event}")
         else:
-            print(f"{Style.BRIGHT + color[op] + op} event {event + Fore.RESET + Style.NORMAL}")
-            
+            print(
+                f"{Style.BRIGHT + color[op] + op} event {event + Fore.RESET + Style.NORMAL}"
+            )
+
         for f, op in sorted(tdiffs.get(event, [])):
             if no_color:
                 print(f"  {op + ' ' + f}")
-            else:            
+            else:
                 print(f"  {color[op] + op + ' ' + f + Fore.RESET}")
         print()
     plan = ", ".join(
@@ -91,17 +93,17 @@ def login(args) -> None:
         if not instance:
             instance = curr_instance
     CONFIG.instance = instance
+    CONFIG.save(update_default_instance=True)
 
     auth = CognitoAuth()
     auth.fetch_new_tokens()
-    CONFIG.save()
     print("Authentication successful.")
 
 
 def plan(args) -> None:
     scowl_dir = args.scowl_dir or CONFIG.scowl_dir
-    client = Client(args.instance)
-    print(f"Comparing '{scowl_dir}' to LIVE on {CONFIG.instance} ({client.tenant})\n")
+    client = Client()
+    print(f"Comparing '{scowl_dir}' to LIVE on {CONFIG.instance} ({client.tenant()})\n")
     branch = _tmp_branch_name()
     client.create_branch_from_dir(scowl_dir, branch)
     try:
@@ -120,7 +122,7 @@ def apply(args):
             print(
                 f"""\nDo you want to perform the above actions?
       Only 'yes' will be accepted to approve.\n"""
-            )            
+            )
         else:
             print(
                 f"""\nDo you want to perform the above actions?
@@ -136,17 +138,17 @@ def apply(args):
             print("Aborting.")
             sys.exit(1)
 
-    Client(args.instance).publish_dir(args.scowl_dir)
+    Client().publish_dir(args.scowl_dir)
     print("Successfully published to LIVE.")
 
 
 def push(args):
-    branch = Client(args.instance).create_branch_from_dir(args.scowl_dir, args.branch)
+    branch = Client().create_branch_from_dir(args.scowl_dir, args.branch)
     print(f"Successfully pushed to branch '{branch}'.")
 
 
 def timeline_list(args):
-    timelines = Client(args.instance).get_timelines().sort_index()
+    timelines = Client().get_timelines().sort_index()
     if args.show:
         print(timelines)
     else:
@@ -154,23 +156,28 @@ def timeline_list(args):
 
 
 def timeline_delete(args):
-    deleted = Client(args.instance).delete_timeline(args.timeline)
+    deleted = Client().delete_timeline(args.timeline)
     print(f"Successfully deleted timeline '{deleted}'.")
 
 
+def timeline_schema(args):
+    scowl = Client().infer_schema_from_timeline(args.timeline)
+    print(scowl)
+
+
 def timeline_show(args):
-    print(Client(args.instance).get_timeline(args.timeline))
+    print(Client().get_timeline(args.timeline))
 
 
 def timeline_upload(args):
     with open(args.jsonl, "r") as f:
         jsonl = f.read()
-        Client(args.instance).create_timeline_from_jsonl(jsonl, args.timeline)
+        Client().create_timeline_from_jsonl(jsonl, args.timeline)
     print(f"Successfully uploaded to timeline '{args.timeline}'.")
 
 
 def branch_list(args):
-    branches = Client(args.instance).get_branches().sort_index()
+    branches = Client().get_branches().sort_index()
     if args.show:
         print(branches)
     else:
@@ -178,25 +185,25 @@ def branch_list(args):
 
 
 def branch_delete(args):
-    deleted = Client(args.instance).delete_branch(args.branch)
+    deleted = Client().delete_branch(args.branch)
     print(f"Successfully deleted branch '{deleted}'.")
 
 
 def branch_show(args):
-    print(Client(args.instance).get_branch(args.branch))
+    print(Client().get_branch(args.branch))
 
 
 def branch_select(args):
     CONFIG.default_branch = args.branch
-    CONFIG.save(update_default_instance=False)
+    print(f"Default branch config updated to '{args.branch}' for {CONFIG.instance}\n")
 
 
 def version(args):
-    print(Client(args.instance).version())
+    print(Client().version())
 
 
 def tenant(args):
-    print(Client(args.instance).tenant)
+    print(Client().tenant())
 
 
 def admin_list_tenants(args):
@@ -252,6 +259,11 @@ TIMELINE_COMMANDS = [
         "help": "delete remote timeline",
         "handler": timeline_delete,
     },
+    {
+        "name": "schema",
+        "help": "display timeline schema as scowl",
+        "handler": timeline_schema,
+    },
     {
         "name": "show",
         "help": "display timeline metadata",
@@ -338,6 +350,8 @@ def main():
 
     parser.add_argument("--debug", action="store_true", help="run in debug mode")
 
+    parser.add_argument("--instance", metavar="URL", help="sumatra instance url")
+
     cmd_parsers = parser.add_subparsers(
         title="commands", metavar="CMD", help="run `sumatra CMD -h` for command help"
     )
@@ -351,13 +365,13 @@ def main():
         if "handler" in cmd:
             p.set_defaults(handler=cmd["handler"])
 
-        p.add_argument("--instance", metavar="URL", help="sumatra instance url")
-
         if cmd["name"] in ["plan", "apply", "push"]:
             p.add_argument("--scowl-dir", metavar="DIR", help="path to scowl files")
-            
+
         if cmd["name"] in ["plan", "apply"]:
-            p.add_argument("--no-color", action="store_true", help="disable color in output")
+            p.add_argument(
+                "--no-color", action="store_true", help="disable color in output"
+            )
 
         if cmd["name"] == "apply":
             p.add_argument(
@@ -396,7 +410,7 @@ def main():
                             "--show", action="store_true", help="display metadata"
                         )
                 if cmd["name"] == "timeline":
-                    if subcmd["name"] in ["delete", "show", "upload"]:
+                    if subcmd["name"] in ["delete", "show", "upload", "schema"]:
                         p.add_argument("timeline", help="remote timeline name")
                     if subcmd["name"] == "list":
                         p.add_argument(
@@ -430,6 +444,9 @@ def main():
         format="%(levelname)s | %(asctime)s | %(name)s | %(message)s",
     )
 
+    if args.instance:
+        CONFIG.instance = args.instance
+
     if hasattr(args, "handler"):
         try:
             args.handler(args)
