@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import os
 import time
 import logging
@@ -44,11 +46,26 @@ def _load_scowl_files(dir: str) -> Dict[str, str]:
 
 
 class Client:
+    """
+    Client to connect to Sumatra GraphQL API
+
+    __Humans:__ First, log in via the CLI: `sumatra login`
+
+    __Bots:__ Set the `SUMATRA_INSTANCE` and `SUMATRA_SDK_KEY` environment variables
+    """
+
     def __init__(
         self,
         instance: Optional[str] = None,
         branch: Optional[str] = None,
     ):
+        """
+        Create connection object.
+
+        Arguments:
+            instance: Sumatra instance url, e.g. `yourco.sumatra.ai`. If unspecified, the your config default will be used.
+            branch: Set default branch. If unspecified, your config default will be used.
+        """
         if instance:
             CONFIG.instance = instance
         if CONFIG.sdk_key:
@@ -58,20 +75,37 @@ class Client:
             auth = CognitoJwtAuth()
             endpoint = CONFIG.console_graphql_url
 
-        self.branch = branch or CONFIG.default_branch
-        self._tenant = None
+        self._branch = branch or CONFIG.default_branch
 
         self._gql_client = python_graphql_client.GraphqlClient(
             auth=auth, endpoint=endpoint
         )
 
     @property
-    def tenant(self):
-        if self._tenant is None:
-            self._tenant = self._fetch_tenant()
-        return self._tenant
+    def instance(self) -> str:
+        """
+        Instance name from client config, e.g. `'yourco.sumatra.ai'`
+        """
+        return CONFIG.instance
 
-    def _fetch_tenant(self):
+    @property
+    def branch(self) -> str:
+        """
+        Default branch name
+        """
+        return self._branch
+
+    @branch.setter
+    def branch(self, branch: str) -> None:
+        self._branch = branch
+
+    def tenant(self) -> str:
+        """
+        Return the assigned tenant name for the connected user.
+
+        Returns:
+            Tenant name
+        """
         logger.debug("Fetching tenant")
         query = """
             query Tenant {
@@ -87,24 +121,30 @@ class Client:
             raise Exception(ret["errors"][0]["message"])
 
         return ret["data"]["tenant"]["key"]
-    
-        
-    def get_branch(self, branch: Optional[str] = None) -> pd.Series:
-        branch = branch or self.branch
+
+    def get_branch(self, branch: Optional[str] = None) -> Dict:
+        """
+        Return metadata about the branch.
+
+        Arguments:
+            branch: Specify a branch other than the client default.
+
+        Returns:
+            Branch metadata
+        """
+        branch = branch or self._branch
         logger.info(f"Getting branch {branch}")
         query = """
             query BranchScowl($id: String!) { 
               branch(id: $id) { id, hash, events, creator, lastUpdated, error } 
             }
         """
-        
-        ret = self._gql_client.execute(
-            query=query, variables={"id": branch}
-        )
-        
+
+        ret = self._gql_client.execute(query=query, variables={"id": branch})
+
         if "errors" in ret:
             raise Exception(f"Error: {ret['errors'][0]['message']}")
-            
+
         branch = ret["data"]["branch"]
         row = {
             "name": branch["id"],
@@ -112,92 +152,116 @@ class Client:
             "update_ts": branch["lastUpdated"],
             "event_types": branch["events"],
         }
-        
+
         if "error" in branch and branch["error"]:
             row["error"] = branch["error"]
-            
-            
+
         return row
-        
-    def clone_branch(self, dest: str, branch: Optional[str] = None) -> str:
-        branch = branch or self.branch
+
+    def clone_branch(self, dest: str, branch: Optional[str] = None) -> None:
+        """
+        Copy branch to another branch name.
+
+        Arguments:
+            dest: Name of branch to be created or overwritten.
+            branch: Specify a source branch other than the client default.
+        """
+        branch = branch or self._branch
         logger.info(f"Cloning branch {branch} to {dest}")
         query = """
             mutation CloneBranch($id: String!, $sourceId: String!) {
                 cloneBranch(id: $id, sourceId: $sourceId) { id, creator, lastUpdated, scowl }
               }
         """
-        
+
         ret = self._gql_client.execute(
             query=query, variables={"id": dest, "sourceId": branch}
         )
-        
+
         if "errors" in ret:
             raise Exception(f"Error: {ret['errors'][0]['message']}")
-        
-        destId = ret["data"]["cloneBranch"]["id"]
-        return destId
-        
-    def put_branch_object(self, key: str, scowl: str, branch: Optional[str] = None) -> str:
-        branch = branch or self.branch
+
+    def _put_branch_object(
+        self, key: str, scowl: str, branch: Optional[str] = None
+    ) -> str:
+        branch = branch or self._branch
         logger.info(f"Putting branch object {key} to branch {branch}")
         query = """
               mutation PutBranchObject($branchId: String!, $key: String!, $scowl: String!) {
                 putBranchObject(branchId: $branchId, key: $key, scowl: $scowl) { key }
               }
         """
-        
+
         ret = self._gql_client.execute(
             query=query, variables={"branchId": branch, "key": key, "scowl": scowl}
         )
-        
+
         if "errors" in ret:
             raise Exception(f"Error: {ret['errors'][0]['message']}")
 
-        key = ret["data"]["putBranchObject"]["key"]
-        return key
-    
-    def create_branch_from_scowl(self, scowl: str, branch: Optional[str] = None) -> str:
-        branch = branch or self.branch
+    def create_branch_from_scowl(
+        self, scowl: str, branch: Optional[str] = None
+    ) -> None:
+        """
+        Create (or overwrite) branch with single file of scowl source code.
+
+        Arguments:
+            scowl: Scowl source code as string.
+            branch: Specify a source branch other than the client default.
+        """
+
+        branch = branch or self._branch
         logger.info(f"Creating branch '{branch}' from scowl")
         try:
             self.delete_branch(branch)
         except:
             pass
-            
-        self.put_branch_object("main.scowl", scowl, branch)
-            
+
+        self._put_branch_object("main.scowl", scowl, branch)
+
         b = self.get_branch(branch)
         if "error" in b:
             raise Exception(b["error"])
-        
-        return b["name"]
-        
+
     def create_branch_from_dir(
         self, scowl_dir: Optional[str] = None, branch: Optional[str] = None
-    ) -> str:
+    ) -> None:
+        """
+        Create (or overwrite) branch with local scowl files.
+
+        Arguments:
+            scowl_dir: Path to local .scowl files.
+            branch: Specify a source branch other than the client default.
+        """
         scowl_dir = scowl_dir or CONFIG.scowl_dir
-        branch = branch or self.branch
+        branch = branch or self._branch
         logger.info(f"Creating branch '{branch}' from dir '{scowl_dir}'")
-        
+
         try:
             self.delete_branch(branch)
         except:
             pass
-        
+
         scowls = _load_scowl_files(scowl_dir)
-        # combine for the time-being until multiple files are supported
+        if not scowls:
+            raise Exception(
+                f"Unable to push local dir. '{scowl_dir}' has no .scowl files."
+            )
+
         for key in scowls:
-            self.put_branch_object(key, scowls[key], branch)
-        
-        
+            self._put_branch_object(key, scowls[key], branch)
+
         b = self.get_branch(branch)
         if "error" in b:
             raise Exception(b["error"])
-        
-        return b["name"]
-        
+
     def publish_dir(self, scowl_dir: Optional[str] = None) -> None:
+        """
+        Create 'main' branch from local scowl dir and promote to LIVE.
+
+        Arguments:
+            scowl_dir: Path to .scowl files. Default: `'.'`
+        """
         scowl_dir = scowl_dir or CONFIG.scowl_dir
         logger.info(f"Publishing dir '{scowl_dir}' to LIVE.")
         branch = "main"
@@ -205,7 +269,13 @@ class Client:
         self.publish_branch(branch)
 
     def publish_branch(self, branch: Optional[str] = None) -> None:
-        branch = branch or self.branch
+        """
+        Promote branch to LIVE.
+
+        Arguments:
+            branch: Specify a branch other than the client default.
+        """
+        branch = branch or self._branch
         logger.info(f"Publishing '{branch}' branch to LIVE.")
         query = """
             mutation PublishBranch($id: String!) {
@@ -222,13 +292,31 @@ class Client:
             )
 
     def publish_scowl(self, scowl: str) -> None:
+        """
+        Create 'main' branch from scowl source and promote to LIVE.
+
+        Arguments:
+            scowl: Scowl source code as string.
+        """
         logger.info("Publishing scowl to LIVE.")
         branch = "main"
         self.create_branch_from_scowl(scowl, branch)
         self.publish_branch(branch)
 
-    def diff_branch_with_live(self, branch: Optional[str] = None) -> Dict:
-        branch = branch or self.branch
+    def diff_branch_with_live(
+        self, branch: Optional[str] = None
+    ) -> Dict[str, List[str]]:
+        """
+        Compare branch to LIVE topology and return diff.
+
+        Arguments:
+            branch: Specify a source branch other than the client default.
+
+        Returns:
+            Events and features added, redefined, and deleted.
+        """
+
+        branch = branch or self._branch
         logger.info(f"Diffing '{branch}' branch against LIVE.")
         query = """
             query Branch($id: String!) {
@@ -257,6 +345,12 @@ class Client:
         return ret["data"]["branch"]["liveDiff"]
 
     def get_branches(self) -> pd.DataFrame:
+        """
+        Return all branches and their metadata.
+
+        Returns:
+            Branch metadata.
+        """
         logger.debug(f"Getting branches")
         query = """
             query BranchList {
@@ -285,7 +379,7 @@ class Client:
             }
             if branch["error"]:
                 row["error"] = branch["error"]
-                
+
             rows.append(row)
         if not rows:
             return pd.DataFrame(columns=["name", "creator", "update_ts", "event_types"])
@@ -295,7 +389,13 @@ class Client:
             "name"
         )
 
-    def get_live_scowl(self):
+    def get_live_scowl(self) -> str:
+        """
+        Return scowl source code for LIVE topology as single cleansed string.
+
+        Returns:
+            Scowl source code as string.
+        """
         query = """
             query LiveScowl {
                 liveBranch { scowl }
@@ -310,8 +410,14 @@ class Client:
         scowl = ret["data"]["liveBranch"]["scowl"]
         return scowl
 
-    def delete_branch(self, branch: Optional[str] = None) -> str:
-        branch = branch or self.branch
+    def delete_branch(self, branch: Optional[str] = None) -> None:
+        """
+        Delete server-side branch
+
+        Arguments:
+            branch: Specify a branch other than the client default.
+        """
+        branch = branch or self._branch
         logger.info(f"Deleting branch '{branch}'.")
         query = """
             mutation DeleteBranch($id: String!) {
@@ -325,10 +431,16 @@ class Client:
 
         if "errors" in ret:
             raise Exception(ret["errors"][0]["message"])
-        return branch
 
-    def get_timelines(self):
-        logger.info(f"")
+    def get_timelines(self) -> pd.DataFrame:
+        """
+        Return all timelines and their metadata.
+
+        Returns:
+            Timeline metadata.
+        """
+
+        logger.debug(f"Getting timelines")
         query = """
             query TimelineList {
                 timelines { id, createUser, createTime, metadata { start, end, count, events }, source, state, error }
@@ -377,13 +489,28 @@ class Client:
         )
 
     def get_timeline(self, timeline: str) -> pd.Series:
+        """
+        Return metadata about the timeline.
+
+        Arguments:
+            timeline: Name of timeline.
+
+        Returns:
+            Timeline metadata.
+        """
         logger.debug(f"Getting timeline '{timeline}'")
         timelines = self.get_timelines()
         if timeline in timelines.index:
             return timelines.loc[timeline]
         raise Exception(f"Timeline '{timeline}' not found.")
 
-    def delete_timeline(self, timeline: str) -> str:
+    def delete_timeline(self, timeline: str) -> None:
+        """
+        Delete timeline
+
+        Arguments:
+            timeline: Name of timeline.
+        """
         logger.info(f"Deleting timeline '{timeline}'.")
         query = """
             mutation DeleteTimeline($id: String!) {
@@ -397,15 +524,66 @@ class Client:
 
         if "errors" in ret:
             raise Exception(ret["errors"][0]["message"])
-        return timeline
 
-    def create_timeline_from_dataframes(self, df_dict, timeline, timestamp_column=None):
+    def infer_schema_from_timeline(self, timeline: str) -> str:
+        """
+        Attempt to infer the paths and data types of all fields in the timeline's
+        input data. Generate the scowl to parse all JSON paths.
+
+        This function helps bootstrap scowl code for new event types, with
+        the expectation that most feature names will need to be modified.
+
+        e.g.
+        ```
+            account_id := $.account.id as int
+            purchase_items_0_amount := $.purchase.items[0].amount as float
+        ```
+
+        Arguments:
+            timeline: Name of timeline.
+
+        Returns:
+            Scowl source code as string.
+        """
+        query = """
+            query TimelineScowl($id: String!) {
+                timeline(id: $id) { id, scowl }
+            }
+        """
+
+        ret = self._gql_client.execute(query=query, variables={"id": timeline})
+
+        if "errors" in ret:
+            raise Exception(ret["errors"][0]["message"])
+        return ret["data"]["timeline"]["scowl"]
+
+    def create_timeline_from_dataframes(
+        self,
+        df_dict: Dict[str, pd.DataFrame],
+        timeline: str,
+        timestamp_column: Optional[str] = None,
+    ) -> None:
+        """
+        Create (or overwrite) timeline from a collection of DataFrames—one per event type.
+
+        Arguments:
+            df_dict: Dictionary from event type name to DataFrame of events.
+            timeline: Timeline name.
+            timestamp_column: Name of timestamp column, default `_time`. If column not present, current timestamp used for all events.
+        """
         jsonl = ""
         for event_type, df in df_dict.items():
             jsonl += self._df_to_jsonl(df, event_type, timestamp_column)
-        return self.create_timeline_from_jsonl(jsonl, timeline)
+        self.create_timeline_from_jsonl(jsonl, timeline)
 
-    def create_timeline_from_jsonl(self, jsonl, timeline):
+    def create_timeline_from_jsonl(self, jsonl: str, timeline: str) -> None:
+        """
+        Create (or overwrite) timeline from JSON events passed in as a string.
+
+        Arguments:
+            jsonl: JSON event data, one JSON dict per line.
+            timeline: Timeline name.
+        """
         query = """
             mutation SaveTimelineMutation($id: String!,
                                           $filename: String!) {
@@ -442,16 +620,20 @@ class Client:
         if "errors" in ret:
             raise Exception(ret["errors"][0]["message"])
 
+        retries = 60
+        delay = 5.0
         retry_count = 0
-        while retry_count < 60:
+        while retry_count < retries:
             tl = self.get_timeline(timeline)
             if tl.status == "processing":
-                time.sleep(5.0)
+                time.sleep(delay)
                 retry_count += 1
             else:
-                break
+                return
 
-        return tl.name
+        raise Exception(
+            f"Timeline processing timed out after {retries * delay} seconds."
+        )
 
     def _df_to_jsonl(self, df, event_type, timestamp_column=None):
         df = df.copy()
@@ -466,11 +648,40 @@ class Client:
             jsonl += "\n"
         return jsonl
 
-    def materialize(self, timeline: str, branch: Optional[str] = None):
+    def materialize(
+        self, timeline: str, branch: Optional[str] = None
+    ) -> Materialization:
+        """
+        Enrich timeline using topology at branch.
+
+        This is the primary function of the SDK.
+
+        Arguments:
+            timeline: Timeline name.
+            branch: Specify a source branch other than the client default.
+
+        Returns:
+            Handle to Materialization job
+        """
+
         return self.materialize_many([timeline], branch)
 
-    def materialize_many(self, timelines: List[str], branch: Optional[str] = None):
-        branch = branch or self.branch
+    def materialize_many(
+        self, timelines: List[str], branch: Optional[str] = None
+    ) -> Materialization:
+        """
+        Enrich collection of timelines using topology at branch. Timelines are merged based on timestamp.
+
+        This is the primary function of the SDK.
+
+        Arguments:
+            timelines: Timeline names.
+            branch: Specify a source branch other than the client default.
+
+        Returns:
+            Handle to Materialization job
+        """
+        branch = branch or self._branch
         query = """
             mutation Materialize($timelines: [String], $branch: String!) {
                 materialize(timelines: $timelines, branch: $branch) { id }
@@ -486,7 +697,13 @@ class Client:
 
         return Materialization(self, ret["data"]["materialize"]["id"])
 
-    def get_models(self):
+    def get_models(self) -> pd.DataFrame:
+        """
+        Return all PMML models and their metadata.
+
+        Returns:
+            Model metadata.
+        """
         query = """
                     query ModelList {
                         models { id, name, version, creator }
@@ -512,7 +729,18 @@ class Client:
             ["creator", "name", "version"], ascending=False
         ).set_index("id")
 
-    def put_model(self, name, version, file):
+    def put_model(self, name: str, version: str, filename: str) -> str:
+        """
+        Upload PMML model from file.
+
+        Arguments:
+            name: Model name, e.g. "churn_predictor".
+            version: Model version id, e.g. "0.8.1".
+            filename: Local PMML file, e.g. "my_model.xml"
+
+        Returns:
+            Server-side model identifier
+        """
         query = """
                     mutation PutModel($name: String!, $version: String!) {
                         putModel (name: $name, version: $version) { id, name, version, uploadUri }
@@ -527,15 +755,21 @@ class Client:
 
         uploadUri = ret["data"]["putModel"]["uploadUri"]
 
-        with open(file, "rb") as f:
-            files = {"file": (file, f)}
+        with open(filename, "rb") as f:
+            files = {"file": (filename, f)}
             http_response = requests.put(uploadUri, files=files)
             if http_response.status_code != 200:
                 raise Exception(http_response.error)
 
         return ret["data"]["putModel"]["id"]
 
-    def version(self):
+    def version(self) -> str:
+        """
+        Return the server-side version number.
+
+        Returns:
+            Version identifier
+        """
         query = """
             query Version {
                 version
@@ -548,7 +782,7 @@ class Client:
 
         return ret["data"]["version"]
 
-    def get_session(self):
+    def _get_session(self):
         query = """
                     query TempCredentials {
                         tenant { credentials }
@@ -568,12 +802,16 @@ class Client:
             aws_session_token=creds["SessionToken"],
         )
 
-    @property
-    def s3_client(self):
-        return self._s3_client
-
 
 class Materialization:
+    """
+    A handle to a server-side materialization job, which enriches a Timeline
+    (or set of Timelines) by running it (them) through a given Topology.
+
+    Objects are not constructed directly. Materializations are returned by methods
+    of the `Client` class.
+    """
+
     def __init__(self, client, id):
         self._client = client
         self.id = id
@@ -609,33 +847,59 @@ class Materialization:
             raise Exception(f"Timed out after {DELAY * RETRIES} seconds")
 
     @property
-    def status(self):
+    def status(self) -> str:
+        """
+        Current status of the job. One of {'processing', 'materialized', 'error'}
+        """
         self._mtr = self._get_materialization()
         return self._mtr["state"]
 
     @property
-    def timelines(self):
+    def timelines(self) -> List[str]:
+        """
+        Timelines materialized by the job.
+        """
         self._wait_for_processing()
         return self._mtr["timelines"]
 
     @property
-    def branch(self):
+    def branch(self) -> str:
+        """
+        Topology branch used for materialization.
+        """
         self._wait_for_processing()
         return self._mtr["branch"]
 
     @property
-    def hash(self):
+    def hash(self) -> str:
+        """
+        Unique hash identifying the job.
+        """
         self._wait_for_processing()
         return self._mtr["hash"]
 
     @property
-    def path(self):
+    def path(self) -> str:
+        """
+        S3 bucket path where results are stored.
+        """
         self._wait_for_processing()
         return self._mtr["path"]
 
-    def get_events(self, event_type, features=[]):
+    def get_events(self, event_type, features=[]) -> pd.DataFrame:
+        """
+        Return enriched events of specified type. Waits if
+        job is still processing.
+
+        Arguments:
+            event_type: Name of event type to fetch.
+            features: Feature names to fetch. By default, fetch all.
+
+        Returns:
+            Enriched events
+        """
         self._wait_for_processing()
-        session = self._client.get_session()
+        session = self._client._get_session()
 
         if not features:
             df = wr.s3.read_parquet(
@@ -656,25 +920,43 @@ class Materialization:
         df = tz_convert_timestamp_columns(df)
         return df.set_index("_id")
 
-    def get_errors(self, event_type, features=[]):
-        self._wait_for_processing()
-        session = self._client.get_session()
+    def get_errors(self, event_type: str, features: List[str] = []) -> pd.DataFrame:
+        """
+        Return event-level materialization errors for specified event type. Waits if
+        job is still processing.
 
-        if not features:
-            df = wr.s3.read_parquet(
-                boto3_session=session,
-                path=f"{self.path}/{event_type}.errors.parquet",
-                ignore_empty=True,
-            )
-        else:
-            cols = ["_id", "_type", "_time"]
-            cols.extend(features)
-            df = wr.s3.read_parquet(
-                boto3_session=session,
-                path=f"{self.path}/{event_type}.errors.parquet",
-                columns=cols,
-                ignore_empty=True,
-            )
+        Arguments:
+            event_type: Name of event type to fetch.
+            features: Feature names to fetch. By default, fetch all.
+
+        Returns:
+            Event-level errors
+        """
+
+        self._wait_for_processing()
+        session = self._client._get_session()
+
+        path = f"{self.path}/{event_type}.errors.parquet"
+        cols = ["_id", "_type", "_time"]
+        cols.extend(features)
+
+        try:
+            if not features:
+                df = wr.s3.read_parquet(
+                    boto3_session=session,
+                    path=path,
+                    ignore_empty=True,
+                )
+            else:
+                df = wr.s3.read_parquet(
+                    boto3_session=session,
+                    path=path,
+                    columns=cols,
+                    ignore_empty=True,
+                )
+        except Exception as e:
+            logger.debug(e)
+            return pd.DataFrame(columns=cols)
 
         df = tz_convert_timestamp_columns(df)
         return df.set_index("_id")
